/*
 * The MIT License
 *
 * Copyright 2012 Massive Dynamics.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package biz.massivedynamics.aurora.runescape;

import biz.massivedynamics.aurora.Aurora;
import biz.massivedynamics.aurora.runescape.cache.VirtualFileSystem;
import biz.massivedynamics.aurora.runescape.network.netty.factories.GamePipelineFactory;
import biz.massivedynamics.aurora.runescape.network.netty.factories.HTTPPipelineFactory;
import biz.massivedynamics.aurora.runescape.network.netty.factories.JagGrabPipelineFactory;
import biz.massivedynamics.aurora.sourcery.managers.generic.SettingsResourceManager;
import biz.massivedynamics.modger.Logger;
import biz.massivedynamics.modger.message.MessageType;
import biz.massivedynamics.sourcery.resource.impl.Settings;
import biz.massivedynamics.versioneer.version.VersionType;
import biz.massivedynamics.versioneer.version.impl.GenericVersion;
import java.net.InetSocketAddress;
import java.util.concurrent.Executors;
import org.jboss.netty.bootstrap.ServerBootstrap;
import org.jboss.netty.channel.ChannelPipelineFactory;
import org.jboss.netty.channel.socket.nio.NioServerSocketChannelFactory;

/**
 * A class that defines the server part of Aurora
 * 
 * The Aurora class itself is merely a utility class, so...
 * 
 * @author Cruz Bishop
 * @version 0.0.1.4
 */
public class Server {
    
    /**
     * The instance of the Aurora server
     * 
     * @since 0.0.0.1
     */
    private static Server instance = new Server();
    
    /**
     * Gets the instance of Aurora's server
     * 
     * @return The instance
     * @since 0.0.0.1
     */
    public static Server getInstance() {
        return instance;
    }
    
    /**
     * Used to NOT construct a new Server class
     * 
     * @since 0.0.0.1
     */
    private Server() {}
    
    /**
     * Gets Aurora's server's version
     * 
     * @return The version
     * @since 0.0.0.1
     */
    public GenericVersion getVersion() {
        return Aurora.getServerVersion();
    }
    
    /**
     * The logger being used in this server
     * 
     * @since 0.0.0.2
     */
    private Logger logger = new Logger("Server");
    
    /**
     * Gets the server's logger
     * 
     * @return The logger
     * @since 0.0.0.2
     */
    public Logger getLogger() {
        return this.logger;
    }
    
    /**
     * The settings
     * 
     * @since 0.0.0.3
     */
    private Settings settings;
    
    /**
     * Gets the settings
     * 
     * @return The settings
     * @since 0.0.0.3
     */
    public Settings getSettings() {
        return this.settings;
    }
    
    /**
     * The cache / VirtualFileSystem
     * 
     * @since 0.0.1.3
     */
    private VirtualFileSystem cache;
    
    /**
     * Gets the cache's virtual file system
     * 
     * @return The virtual file system
     * @since 0.0.1.3
     */
    public VirtualFileSystem getCache() {
        return this.cache;
    }
    
    /**
     * Starts the server
     * 
     * @since 0.0.0.1
     */
    public void start() {
        
        //Show debugging information for non-stable versions
        if (!Aurora.getServerVersion().getType().equals(VersionType.STABLE)) {
            MessageType.DEBUG.unfilter();
        }
        
        //Load the settings
        this.settings = new SettingsResourceManager().load();
        
        //Log a trace message
        logger.submitTrace("Server startup called");
        
        //Display the loading log
        logger.submitInformation(settings.getSetting("serverName", "Aurora") + " version " + Aurora.getServerVersion() + " is loading");
        
        //And print some debugging information (Client version)
        logger.submitDebug("This server supports version " + Aurora.getClientVersion() + " clients");
        
        logger.submitTrace("Loading cache VirtualFileSystem");
        
        //Set up the virtual file system
        this.cache = new VirtualFileSystem();
        
        this.cache.start();
        
        logger.submitTrace("Starting server bootstraps");
        
        //Start the HTTP and JagGrab bootstraps
        this.addBootstrap("HTTP", (Integer) this.settings.getSetting("httpPort", 80), new HTTPPipelineFactory());
        
        this.addBootstrap("JagGrab", (Integer) this.settings.getSetting("jaggrabPort", 43595), new JagGrabPipelineFactory());
        
        this.addBootstrap("Game", (Integer) this.settings.getSetting("gamePort", 43594), new GamePipelineFactory());
        
        //TODO Add a shutdown hook for the following
        new SettingsResourceManager().save(this.settings);
        
    }
    
    /**
     * Adds a new bootstrap and binds it
     * 
     * @param name The bootstrap's name
     * @param port The port to listen to
     * @param factory The channel pipeline factory to use
     * @since 0.0.1.2
     */
    private void addBootstrap(String name, int port, ChannelPipelineFactory factory) {
        
        //Set up a new bootstrap
        ServerBootstrap bootstrap = new ServerBootstrap(
            new NioServerSocketChannelFactory(
                Executors.newCachedThreadPool(),
                Executors.newCachedThreadPool()
            ));
        
        //Set the pipeline factory
        bootstrap.setPipelineFactory(factory);
        
        try {
            //Bind the bootstrap
            bootstrap.bind(new InetSocketAddress(port));
            
            this.logger.submitMessage(name + " bootstrap bound and started on port " + port);
        } catch (Exception exception) {
            this.logger.submitWarning(name + " bootstrap not started (port " + port + ")", exception);
        }
        
    }
    
}
