/*
 * The MIT License
 *
 * Copyright 2012 Massive Dynamics.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package biz.massivedynamics.aurora.runescape.cache;

import java.io.IOException;
import java.nio.channels.SeekableByteChannel;

/**
 * A index file
 *
 * @author Cruz Bishop
 * @version 1.0.0.2
 */
public class IndexFile {
    
    /**
     * Creates a new index file
     * 
     * @param id The ID to use (0 - 4)
     * @param channel The channel to use
     * @since 1.0.0.0
     */
    public IndexFile(int id, SeekableByteChannel channel) {
        this.id = id;
        this.channel = channel;
    }
    
    /**
     * The ID of this index file
     * 
     * @since 1.0.0.0
     */
    private int id;
    
    /**
     * Gets the id of this index file
     * 
     * @return The ID
     * @since 1.0.0.0
     */
    public int getId() {
        return this.id;
    }
    
    /**
     * The seekable byte channel this file contains
     * 
     * @since 1.0.0.0
     */
    private SeekableByteChannel channel;
    
    /**
     * Gets the seekable byte channel
     * 
     * @return The seekable byte channel
     * @since 1.0.0.0
     */
    public SeekableByteChannel getChannel() {
        return this.channel;
    }
    
    /**
     * Gets the size of this file
     * 
     * @return The size, or zero if the file cannot be accessed
     * @since 1.0.0.0
     */
    public long getSize() {
        try {
            return this.channel.size();
        } catch (IOException ex) {
            return 0;
        }
    }
    
    /**
     * Gets the count of entries handled by this index file
     * 
     * @return The entry count
     * @since 1.0.0.1
     */
    public long getEntryCount() {
        return this.getSize() / 6;
    }
    
    /**
     * Gets a virtual file index
     * 
     * @param entryID The entry ID to look for
     * @return The virtual file index
     * @throws IOException Any exceptions that may occur
     * @since 1.0.0.2
     */
    public VirtualFileIndex getVirtualFileIndex(int entryID) throws IOException {
        return new VirtualFileIndex(this, entryID);
    }
    
}
