/*
 * The MIT License
 *
 * Copyright 2012 Massive Dynamics.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package biz.massivedynamics.aurora.runescape.cache;

import biz.massivedynamics.aurora.runescape.Server;
import biz.massivedynamics.modger.Logger;
import java.io.IOException;
import java.nio.ByteBuffer;

/**
 * A virtual file index that points to a file in the main_file_cache.dat file
 *
 * @author Cruz Bishop
 * @version 1.1.0.0
 */
public class VirtualFileIndex {
    
    /**
     * The logger used in this class
     * 
     * @since 1.0.0.0
     */
    private Logger logger = new Logger("VirtualFileIndex");
    
    /**
     * Loads a virtual file index
     * 
     * @param file The IndexFile to use
     * @param entryID The entry to search for
     * @since 1.0.0.0
     */
    public VirtualFileIndex(IndexFile file, int entryID) throws IOException {
        //First, check the entry ID
        if (entryID < 0 || entryID > file.getEntryCount()) {
            throw new IOException("Invalid entry ID: " + entryID);
        }
        
        //Set up a new directly allocated byte buffer
        ByteBuffer buffer = ByteBuffer.allocateDirect(6);
        
        //First, synchronize the index file so we don't go corrupting other requests
        synchronized(file.getChannel()) {
            //Position the pointer at entryID * 6
            //Since each entry is six bytes long
            file.getChannel().position(entryID * 6);
            //And fill the buffer
            file.getChannel().read(buffer);
        }
        
        //Flip the buffer
        buffer.flip();
        
        //Don't need to check the capacity of the buffer, so skip straight to the loading
        
        //First, get the reported size
        this.size = ((buffer.get() & 0xff) << 16) 
                | ((buffer.get() & 0xff) << 8) 
                | (buffer.get() & 0xff);
        
        //Then get the first block
        this.firstBlock = ((buffer.get() & 0xff) << 16)
                | ((buffer.get() & 0xff) << 8) 
                | (buffer.get() & 0xff);
        
        this.file = file;
        this.entryID = entryID;
        
        logger.submitTrace("VirtualFileIndex: " + this.size + ":" + this.firstBlock);
        
    }
    
    /**
     * The index file associated with this file index
     * 
     * @since 1.1.0.0
     */
    private IndexFile file;
    
    /**
     * Gets the file associated with this file index
     * 
     * @return The file
     * @since 1.1.0.0
     */
    public IndexFile getFile() {
        return this.file;
    }
    
    /**
     * The entry ID
     * 
     * @since 1.1.0.0
     */
    private int entryID;
    
    /**
     * Gets the entry ID
     * 
     * @return The entry ID
     * @since 1.0.0.0
     */
    public int getEntryID() {
        return this.entryID;
    }
    
    /**
     * The size specified by the cache
     * 
     * @since 1.0.0.0
     */
    private int size;
    
    /**
     * Gets the size specified by the cache
     * 
     * @return The size
     * @since 1.0.0.0
     */
    public int getSize() {
        return this.size;
    }
    
    /**
     * The block number of the first block in the cache
     * 
     * @since 1.0.0.0
     */
    private int firstBlock;
    
    /**
     * Gets the block number of the first block in the cache
     * 
     * @return The first block
     * @since 1.0.0.0
     */
    public int getFirstBlock() {
        return this.firstBlock;
    }
    
    /**
     * Gets the data associated with this virtual file index
     * 
     * @return The data
     * @since 1.1.0.0
     * @throws IOException Some I/O error occurred
     */
    public ByteBuffer getData() throws IOException {
        //Allocate a byte buffer that is as large as this file index
        ByteBuffer buffer = ByteBuffer.allocateDirect(this.getSize());
        
        //Calculate the starting position
        long pointer = (long) this.firstBlock * 520L;
        
        //Start a counter to show the number of bytes read
        int bytesRead = 0;
        
        //Get the total number of blocks to read
        int totalBlocks = this.getSize() / 512; //TODO Move this to getBlockCount()
        
        //Do a quick sanity check
        if (this.getSize() % 512 != 0) {
            totalBlocks++;
        }
        
        for (int blockNumber = 0; blockNumber < totalBlocks; blockNumber++) {
            //Read the header first
            ByteBuffer header = ByteBuffer.allocate(8);
            //Synchronize with the data file
            synchronized (Server.getInstance().getCache().getDataFile()) {
                //Position the pointer
                Server.getInstance().getCache().getDataFile().position(pointer);
                //And read the header
                Server.getInstance().getCache().getDataFile().read(header);
            }
            
            pointer += 8;
            
            //Get the information for the next block
            
            //Get the next entry
            int nextEntry = ((header.get(0) & 0xFF) << 8) | (header.get(1) & 0xff);
            //Get the current chunk
            int currentChunk = ((header.get(2) & 0xff) << 8) | (header.get(3) & 0xff);
            //Get the next block
            int nextBlock = ((header.get(4) & 0xff) << 16) 
                    | ((header.get(5) & 0xff) << 8)
                    | (header.get(6) & 0xff);
            
            //Get the next index
            int nextIndex = header.get(7) & 0xff;
            
            //Compare the chunks
            if (blockNumber != currentChunk) {
                throw new IOException("Chunk ID mismatch (" + blockNumber +
                        " : " + currentChunk + ")");
            }
            
            //Calculate the size of the rest of the chunk
            int chunkSize = this.getSize() - bytesRead;
            
            //Do a sanity check
            if (chunkSize > 512) {
                chunkSize = 512;
            }
            
            //Allocate a buffer for the chunk
            ByteBuffer chunk = ByteBuffer.allocate(chunkSize);
            
            //Synchronize, like at the ballet
            synchronized (Server.getInstance().getCache().getDataFile()) {
                //Position the pointer
                Server.getInstance().getCache().getDataFile().position(pointer);
                //And read the chunk
                Server.getInstance().getCache().getDataFile().read(chunk);
            }
            
            //Flip the chunk
            chunk.flip();
            
            //Add the chunk to the data
            buffer.put(chunk);
            
            //Increment the bytes read
            bytesRead += chunkSize;
            
            //Increment the pointer
            pointer = (long) nextBlock * 520L;
            
            //See if we still have data to read
            if (this.getSize() > bytesRead) {
                //Check the next index
                if (nextIndex != this.file.getId() + 1) {
                    throw new IOException("File index mismatch");
                }
                
                //Check the entry
                if (nextEntry != this.entryID) {
                    throw new IOException("Entry Id mismatch");
                }
            }
        }
        
        //Flip the buffer
        buffer.flip();
        
        //And return the buffer
        return buffer;
    }
    
}
