/*
 * The MIT License
 *
 * Copyright 2012 Massive Dynamics.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package biz.massivedynamics.aurora.runescape.cache;

import biz.massivedynamics.aurora.runescape.cache.exceptions.IndexFileNotFoundException;
import biz.massivedynamics.modger.Logger;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SeekableByteChannel;
import java.nio.file.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.zip.CRC32;

/**
 * A virtual file system used by RuneScape's cache
 * Is not represented as a NIO FileSystem because of limitations.
 *
 * @author Cruz Bishop
 * @version 0.0.1.0
 */
public final class VirtualFileSystem {
    
    /**
     * The logger
     * 
     * @since 0.0.0.4
     */
    private Logger logger = new Logger("VirtualFileSystem");
    
    /**
     * Creates a new virtual file system for use by a cache
     * 
     * @since 0.0.0.4
     */
    public VirtualFileSystem() {
    }
    
    /**
     * Starts the virtual file system
     * 
     * @since 0.0.1.0
     */
    public void start() {        
        //Check to see if the cache doesn't exist
        if (!this.doesCacheExist()) {
            logger.submitError("The cache does not exist!",
                    "Aurora cannot start without a cache.",
                    "Please place a 377 cache in Data/Cache/");
            
            System.exit(1);
        } else {
            logger.submitDebug("Cache detected");
        }
        //Set a directory stream
        DirectoryStream<Path> directory = null;
        try {
            //Get a directory stream
             directory = Files.newDirectoryStream(this.getCacheDirectory());
            
            //Iterate over the paths
            for (Path indexFilePath : directory) {
                //Get the file name
                Path indexFileName = indexFilePath.getFileName();
                //See if the file name is an index file
                if (indexFileName.toString().startsWith("main_file_cache.idx")) {
                    //Get the index number
                    Integer indexNumber = Character.getNumericValue(indexFileName.toString().charAt(indexFileName.toString().length() - 1));
                    
                    logger.submitTrace("Index " + indexNumber + " found");
                    
                    //Add the index file path
                    this.indexFiles.put(indexNumber, new IndexFile(indexNumber, Files.newByteChannel(indexFilePath, StandardOpenOption.READ)));
                } else continue;
            }
        } catch (IOException ex) {
            logger.submitError("Could not read directory stream", ex);
        } finally {
            try {
                if (directory != null)
                    directory.close();
                } catch (Exception e) {
                    logger.submitWarning("Unknown exception", e);
                }
            
            logger.submitDebug("Loaded " + this.indexFiles.size() + " index files");
        }
        
        //Index files are loaded, begin loading the data file
        try {
            //Open a read-only data file
            this.dataFile = Files.newByteChannel(this.getCacheDirectory().resolve("main_file_cache.dat"), StandardOpenOption.READ);
            //Log the fact that it was opened
            logger.submitDebug("Loaded the data file (" + (this.dataFile.size() / 1024 / 1024) +"MB)");
        } catch (IOException exception) {
            //Submit an error message
            logger.submitError("Could not access data file", exception);
            //And exit
            System.exit(1);
        }
        
        //Start verifying the index files
        for (IndexFile indexFile : this.indexFiles.values()) {
            //See if the file is empty or, somehow, a negative size
            if (indexFile.getSize() <= 0) {
                //Invalid index file
                logger.submitError("Index file " + indexFile.getId() + " is empty?");
                System.exit(1);
            }
            //Check the length is a multiple of six
            if ((indexFile.getSize() % 6) != 0) {
                //Bah, invalid index file
                logger.submitError("Invalid size for index file " + indexFile.getId(),
                        "Must be a multiple of six.",
                        "Size: " + indexFile.getSize());
                System.exit(1);
            }
        }
        
        //Verify the main index file
        if (this.indexFiles.get(0).getEntryCount() != 9) {
            logger.submitWarning("Main index file (idx0) has an unexpected number of indices (" 
                    + indexFiles.get(0).getEntryCount() + ')');
        }
        
        logger.submitDebug("Index files verified");
                
        //Cache is loaded. Report back
        logger.submitMessage("Cache loaded");
        try {
            //Generate the CRC table
            this.generateCRCTable();
            logger.submitInformation("CRC table generated");
        } catch (IOException ex) {
            logger.submitError("Could not generate CRC table", ex);
            System.exit(1);
        }
    }
    
    /**
     * The seekable byte channel that represents the data file
     * 
     * @since 0.0.0.5
     */
    private SeekableByteChannel dataFile;
    
    /**
     * Gets the data file
     * 
     * @return The data file
     * @since 0.0.0.5
     */
    public SeekableByteChannel getDataFile() {
        return this.dataFile;
    }
    
    /**
     * A hash map of index files
     * 
     * @since 0.0.0.4
     */
    private ConcurrentHashMap<Integer, IndexFile> indexFiles = new ConcurrentHashMap<>();
    
    /**
     * Gets an index file
     * 
     * @param index The index to search
     * @return The index file
     * @throws IndexFileNotFoundException Index file not found
     * @since 0.0.0.4
     */
    public IndexFile getIndexFile(Integer index) throws IndexFileNotFoundException {
        if (indexFiles.containsKey(index)) {
            return indexFiles.get(index);
        } else throw new IndexFileNotFoundException("Index " + index + " not found");
    }
    
    /**
     * Gets the cache directory
     * 
     * @return The cache directory
     * @since 0.0.0.1
     */
    public Path getCacheDirectory() {
        return Paths.get("Data", "Cache");
    }
    
    /**
     * Checks to see if the cache (data file) exists
     * 
     * @return True if found, otherwise false
     * @since 0.0.0.1
     */
    public boolean doesCacheExist() {
        return Files.exists(getCacheDirectory().resolve("main_file_cache.dat"));
    }
    
    /**
     * Checks to see if an index file exists on the file system
     * 
     * @param index The index to check     * 
     * @return True if found, otherwise false
     * @since 0.0.0.3
     */
    public boolean doesIndexFileExist(int index) {
        return Files.exists(getCacheDirectory().resolve("main_file_cache.idx" + index));
    }
    
    /**
     * Generates the CRC table
     * 
     * @since 0.0.1.0
     */
    private void generateCRCTable() throws IOException {
        //Get the number of archives
        int archives = (int) this.indexFiles.get(0).getEntryCount();
        
        //Define the hash
        int hash = 1234;
        
        //Define the CRCs
        int[] crcs = new int[archives];
        
        //Calculate the CRCs
        CRC32 crc = new CRC32();
        
        //Loop
        for (int entryID = 1; entryID < crcs.length; entryID++) {
            crc.reset();
            
            //Set up a temporary buffer
            ByteBuffer tempBuffer = this.indexFiles.get(0).getVirtualFileIndex(entryID).getData();
            
            //Get a byte array
            byte[] bytes = new byte[tempBuffer.remaining()];
            
            //And "get" the bytes
            tempBuffer.get(bytes, 0, bytes.length);
            
            //Update the CRC key
            crc.update(bytes, 0, bytes.length);
            
            //And the array
            crcs[entryID] = (int) crc.getValue();
        }
        
        //Allocate a new buffer
        ByteBuffer buffer = ByteBuffer.allocateDirect(crcs.length * 4 + 4);
        
        //Loop through each CRC key
        for (int id = 0; id < crcs.length; id++) {
            //Change the hash
            hash = (hash << 1) + crcs[id];
            //And put the key
            buffer.putInt(crcs[id]);
        }
        
        //Put the hash in the buffer
        buffer.putInt(hash);
        
        //and flip
        buffer.flip();
        
        //Finally change the table
        this.crcTable = buffer.slice();
    }
    
    /**
     * The CRC table
     * 
     * @since 0.0.1.0
     */
    private ByteBuffer crcTable;
    
    /**
     * Gets the CRC table
     * 
     * @return The CRC table as a byte buffer
     * @since 0.0.1.0
     */
    public ByteBuffer getCRCTable() {
        return this.crcTable.slice();
    }
    
}
