/*
 * The MIT License
 *
 * Copyright 2012 Massive Dynamics.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package biz.massivedynamics.aurora.runescape.network.netty.handlers;

import biz.massivedynamics.aurora.runescape.network.netty.updateserver.RequestHandler;
import biz.massivedynamics.aurora.runescape.network.netty.updateserver.jaggrab.JagGrabRequest;
import biz.massivedynamics.aurora.runescape.network.netty.updateserver.jaggrab.JagGrabResponse;
import biz.massivedynamics.modger.Logger;
import java.nio.ByteBuffer;
import org.jboss.netty.buffer.ChannelBuffers;
import org.jboss.netty.channel.ChannelHandlerContext;
import org.jboss.netty.channel.MessageEvent;
import org.jboss.netty.handler.codec.http.HttpRequest;
import org.jboss.netty.handler.timeout.IdleStateAwareChannelUpstreamHandler;
import org.jboss.netty.handler.timeout.IdleStateEvent;

/**
 * An idle-aware channel handler for the update server
 *
 * @author Cruz Bishop
 * @version 0.1.0.0
 */
public class UpdateServerHandler extends IdleStateAwareChannelUpstreamHandler {

    /**
     * The logger for this class
     * 
     * @since 0.1.0.0
     */
    private Logger logger = new Logger("Update Server");
    
    /**
     * Called when the channel is idle
     * 
     * @param ctx The channel handler context
     * @param e The idle state event
     * @throws Exception Any exceptions that are thrown
     * @since 0.1.0.0
     */
    @Override
    public void channelIdle(ChannelHandlerContext ctx, IdleStateEvent e) throws Exception {
        //Calculate how long the idle is
        long time = System.currentTimeMillis() - e.getLastActivityTimeMillis();
        //Submit a trace log
        logger.submitTrace("Channel became idle for " + time + " at " + e.getChannel().getRemoteAddress().toString());
        //And close the channel
        e.getChannel().close();
    }

    /**
     * Called when a message is received
     * 
     * @param ctx The channel handler context
     * @param e The message event
     * @throws Exception Any exceptions that are thrown
     * @since 0.1.0.0
     */
    @Override
    public void messageReceived(ChannelHandlerContext ctx, MessageEvent e) throws Exception {
        //Get the object received
        Object message = e.getMessage();
        
        //See if it's a HTTP request
        if (message instanceof HttpRequest) {
            //Get a request
            HttpRequest request = (HttpRequest) message;
            logger.submitInformation("HTTP request for " + request.getUri());
        } else if (message instanceof JagGrabRequest) {
            //It's JagGrab! D:
            //Get a request
            JagGrabRequest request = (JagGrabRequest) message;
            
            //Set up the data
            ByteBuffer data = RequestHandler.get(request.getPath());
            
            if (data != null) {
                e.getChannel().write(new JagGrabResponse(ChannelBuffers.wrappedBuffer(data)));
            } else {
                e.getChannel().close();
                logger.submitWarning("Unknown JagGrab request: " + request.getPath());
            }
        } else {
            logger.submitWarning("Unknown message type: " + message.getClass().getName());
        }
    }
    
    
    
}
